//Copyright (c) 2003, The Institute for Genomic Research (TIGR), Rockville,
// Maryland, U.S.A.  All rights reserved.

// This is a GHMM of GlimmerM 
// mpertea@tigr.org

#include  "delcher.h"
#include  "gene.h"
#include "oc1.h"

//#define MESSAGE

#define log2(x) (log(x)/0.693147181)

// ********************* CONSTANT DECLARATIONS **********************
const int  DEFAULT_MIN_GENE_LEN = 60;

const int  MODEL_LEN = 12;
const int  SIMPLE_MODEL_LEN = 6;
const int  ALPHABET_SIZE = 4;
const int  MAX_NAME_LEN = 256;
const double  MAX_LOG_DIFF = -20.0;
const int NSTATES = 19; // total number of states
const int NCSTATES = 7; // number of nocoding states

#include  "context.h"
#include "icm.h"

const unsigned int  OK = 0x0;

// ******************** STRUCTURE DEFINITIONS ***********************

struct SpliceSite 
{
  int type;  // atg=0, gt=1, ag=2, stop=3;
  double score;
};

struct exon
{
  long int start;
  long int stop;
  int type; // 0=init+ 1=internal+ 2=terminal+ 3=singl+ 4=init- 5=internal- 6=terminal- 7=single-
  int lphase;
  int rphase;
  double score;
  struct exon *prev;
  int exon_no;
};

struct Site
{
  double score;
  struct exon *prevex;
};

struct List
{
  exon *predexon;
  List *link;
};

// ******************** FUNCTION DECLARATIONS ***********************

void  Process_Options  (int, char * []);
void  Read_Coding_Model(char *);
tModel * Read_NonCoding_Model(char *);
int  Is_Acceptor  (const int *, double *, double, int);
int  Is_Donor  (const int *, double *, double, int);
int  Is_Atg  (const int *, double *,double);
int  Is_Stop  (const int *, double *,double);
void DP(long int start, long int stop);
int check_for_stops(long int prevstop,long int newstart,int prevrphase);
void printexon(exon *curexon,int *geneno,int *exonno);
double ifhexamer(long int start, long int stop);
int basetoint(char c, int dir);
double score_signalp(long int start, long int stop);
int stopcodon(int frame,int phase,int stop, int start);
int addexon(int number);

// ********************* VARIABLE DECLARATIONS **********************

char  * Data;
long int Data_Len;
char TRAIN_DIR[500]="";
long int  Min_Gene_Len = DEFAULT_MIN_GENE_LEN;
//double  Ch_Ct [ALPHABET_SIZE] = {0.0};   // Count of number of occurrences of acgt's in entire genome
//double Ch_All[ALPHABET_SIZE]={0.0};
int Use_Start_Site = 1;
int Use_Stop_Site = 1;
double Acc_Thr = -99;
double Don_Thr = -99;
double Atg_Thr = -99;
double Stop_Thr = -99;
int Use_Filter = 0;
char CODING_FILE[500]="";
char NONCODING_FILE[500]="";
char SIGNALP_FILE[500]="";
int ifSP=0;
int  Use_Independent = TRUE;
//int Use_Exon_Count = TRUE; // default
int Use_Exon_Count = FALSE; // default
int SignalP[32][4];
int boostexon=5;  // 5 for arab , 7 for aspg 
double boostdistr=0; // 0 for arab & aspg
double snglfactor=1; // 1 for arab & aspg
int ExNo;
tModel * NMODEL;

// States

enum state {INTERG,I0PLUS,I1PLUS,I2PLUS,I0MINUS,I1MINUS,I2MINUS,ESGLPLUS,EINPLUS,E0PLUS,E1PLUS,E2PLUS,ETRPLUS,ESGLMINUS,EINMINUS,E0MINUS,E1MINUS,E2MINUS,ETRMINUS};
double Trans[NSTATES][NSTATES];
char EXON_DISTR[500]="";
char EXONNO_DISTR[500]="";
double *LenDistr[6]; // 0=initial, 1=internal, 2=terminal, 3=single, 4=intron, 5=intergenic
double *ExDistr;
double *ExTail;
int LenNo[4]; // remember it just for exons
Site *Sites[NCSTATES];

// ********************* DECISION TREES DECLARATIONS *****************

#define NO_OF_VAR 2 	/* number of variables for the decision trees */
#define NO_OF_TREES 10     /* number of decision trees  */

char TREENAMES[500]="";

/* this file contains freq's of all in-frame hexamers in all
   training sequences. */
char IF_6MER_TRAIN[500]="";        // "exon.hexfreq"

/* this file contains the freq's of all hexamers in all training
   sequences, including exons, introns, and intergenic DNA */
char TRAIN_6MERS[500]="";          // "all.hexfreq"

int no_of_trees = NO_OF_TREES;
int no_of_dimensions = NO_OF_VAR;
int no_of_categories = 2;
struct tree_node **exonroots = NULL;
struct tree_node **startroots = NULL;
struct tree_node **lastroots = NULL;
struct tree_node **snglroots = NULL;

struct tree_node *atgroot = NULL;
struct tree_node *tagroot = NULL;
struct tree_node *accroot = NULL;
struct tree_node *donroot = NULL;


double *HexamerVec;    /* 4096 vector of in-frame hexamer frequencies */
void classify(POINT *,struct tree_node **);
void classifysgl(POINT *,struct tree_node *);
extern struct tree_node *read_tree(char *);
void decision_tree();
int get_if_hexamer(char *,char *,int,double *);


// ************************** Scoring ***************************

#define FORWARD 0
#define REVERSE 1

const double NO_SCORE=-99;

double *Cod[6];     // scoring in six frames; these are log odds probabilities of coding vs. noncoding
SpliceSite *SS[2];  // all splice sites on both strands: atg=0, gt=1, ag=2, stop=3;
                    // splice site position is always on first and last base of signal 
                    // for forward and reverse strand respectively

exon *scoreinitexon(long int start,long int stop, int *prevstate, int *nextstate);
exon *scoretermexon(long int start,long int stop, int *prevstate, int *nextstate);
exon *scoreinternexon(long int start,long int stop, int rphase, int *prevstate, int *nextstate);
exon *scoresglexon(long int start,long int stop, int *prevstate, int *nextstate);


// ************************** Main PROGRAM ***************************

main  (int argc, char * argv [])
{
  FILE  * fp;
  char  File_Name [MAX_LINE], Name [MAX_LINE], Line [MAX_LINE], Param[MAX_LINE];
  float readval,cgperc;
  long int acgt;
  int istacc=0, istdon=0, isatg=0, isstop=0;
  double score, minterg, m5utr=0, m3utr=0, mintron, i5utr, i3utr, iinterg,iintron, iin[3];
  double noncoding,coding;
  int ret,model;
  long int i,j,k,Input_Size;
  int indexhex, val, maxval;
  int B[200];
  double maxscore;
  exon *curexon;
  List *PredExon, *CurExon;
  int geneno,exonno;

  // ---------------------------- OPTIONS ------------------------------

  if  (argc < 2) {
    fprintf (stderr,
	     "USAGE:  %s <genome-file> [options] \n",
	     argv [0]);
    exit (-1);
  }

  Process_Options (argc, argv);   // Set global variables to reflect status of
                                  // command-line options.

  // --------------------------- READ DATA ------------------------------

  fp = File_Open (argv [1], "r");
  
  Data = (char *) Safe_malloc (INIT_SIZE);
  Input_Size = INIT_SIZE;// Read entire genome into  Data [1 .. ]
  // it seams that if there is not enough space, the space is increased
  Read_String (fp, Data, Input_Size, Name, FALSE);   
  
  fclose (fp);
  
  Data[0]='n';

  Data_Len = strlen (Data + 1);

  printf("GlimmerHMM (Version 1.0)\n");    
  printf("Sequence name: %s\n",Name); 

  printf("Sequence length: %ld bp\n",Data_Len);

  cgperc=0;
  acgt=0;

  for  (i = 1;  i <= Data_Len;  i ++) {

    // Converts all characters to  acgt
    Data [i] = Filter (tolower (Data [i]));

    switch  (Data [i]) {
    case  'c' :
    case  'g' :
      cgperc++;
    case  'a' :
    case  't' :
      acgt++;
      break;
    }

  }

  cgperc/=acgt;
  cgperc*=100;

  printf("Sequence C+G%: %f\n",cgperc);

  // ----------------------------------
  //         read config_file
  // ----------------------------------

  strcpy(File_Name,TRAIN_DIR);
  strcat(File_Name,"config.file"); 
  fp = fopen (File_Name, "r");
  if  (fp == NULL)
    {
      fprintf (stderr, "ERROR:  Unable to open file %s\n", File_Name);
      exit (0);
    }

  File_Name[0]='\0';

  while(fgets (Line, MAX_LINE, fp) != NULL) {
    sscanf(Line,"%*s %*s %f %s",&readval,File_Name);
    if(cgperc<=readval) { break; }
  }

  fclose(fp);

  assert(File_Name[0] != '\0');

  strcpy(Line,TRAIN_DIR);
  strcat(Line,File_Name);

  fp = fopen (Line, "r");
  if  (fp == NULL)
    {
      fprintf (stderr, "ERROR:  Unable to open file %s\n", File_Name);
      exit (0);
    }
  while(fgets (Line, MAX_LINE, fp) != NULL) {
    sscanf(Line,"%s %s", Param, File_Name);
    if(strcmp (Param,"coding_model_file") == 0 ) {
      strcpy(CODING_FILE,File_Name); 
      continue;
    }
    if(strcmp (Param,"noncoding_model_file")==0) {
      strcpy(NONCODING_FILE,File_Name); 
      continue;
    }
    if(strcmp (Param,"acceptor_MDD_tree")==0) {
      istacc=atoi(File_Name);
      assert(istacc == 0 || istacc ==1);
      continue;
    }
    if(strcmp (Param,"donor_MDD_tree")==0) {
      istdon=atoi(File_Name);
      assert(istdon == 0 || istdon ==1);
      continue;
    }
    if(strcmp (Param,"acceptor_threshold")==0) {
      if(Acc_Thr==-99) {
	score = strtod (File_Name,NULL);
	if  (errno == ERANGE)
	  fprintf (stderr, "ERROR:  Bad acceptor threshold score in config_file: %s\n", File_Name);
	else Acc_Thr=score;
      }
      continue;
    }
    if(strcmp (Param,"donor_threshold")==0) {
      if(Don_Thr==-99) {
	score = strtod (File_Name,NULL);
	if  (errno == ERANGE)
	  fprintf (stderr, "ERROR:  Bad donor threshold score in config_file: %s\n", File_Name);
	else Don_Thr=score;
      }
      continue;
    }
    if(strcmp (Param,"ATG_detection")==0) {
      isatg=atoi(File_Name);
      assert(isatg == 0 || isatg ==1);
      continue;
    }
    if(strcmp (Param,"ATG_threshold")==0) {
      if(Atg_Thr==-99) {
	score = strtod (File_Name,NULL);
	if  (errno == ERANGE)
	  fprintf (stderr, "ERROR:  Bad atg threshold score in config_file: %s\n", File_Name);
	else Atg_Thr=score;
      }
      continue;
    }
    if(strcmp (Param,"Stop_detection")==0) {
      isstop=atoi(File_Name);
      assert(isstop == 0 || isstop ==1);
      continue;
    }
    if(strcmp (Param,"Stop_threshold")==0) {
      if(Stop_Thr==-99) {
	score = strtod (File_Name,NULL);
	if  (errno == ERANGE)
	  fprintf (stderr, "ERROR:  Bad stop threshold score in config_file: %s\n", File_Name);
	else Stop_Thr=score;
      }
      continue;
    }
    if(strcmp (Param,"DT_filename")==0) {
      strcpy(TREENAMES,File_Name); 
      continue;
    }
    if(strcmp (Param,"InFrameHexamerFile")==0) {
      strcpy(IF_6MER_TRAIN,File_Name); 
      continue;
    }
    if(strcmp (Param,"AllHexamerFile")==0) {
      strcpy(TRAIN_6MERS,File_Name); 
      continue;
    }
    if(strcmp (Param,"Init5'UTR")==0) {
      i5utr = strtod (File_Name,NULL);
      if  (errno == ERANGE) {
	fprintf (stderr, "ERROR:  Bad init 5'UTR parameter in config_file: %s\n", File_Name);
	exit(0);
      }
      continue;
    }
    if(strcmp (Param,"Init3'UTR")==0) {
      i3utr = strtod (File_Name,NULL);
      if  (errno == ERANGE) {
	fprintf (stderr, "ERROR:  Bad init 3'UTR parameter in config_file: %s\n", File_Name);
	exit(0);
      }
      continue;
    }
    if(strcmp (Param,"InitInterg")==0) {
      iinterg = strtod (File_Name,NULL);
      if  (errno == ERANGE) {
	fprintf (stderr, "ERROR:  Bad init intergenic parameter in config_file: %s\n", File_Name);
	exit(0);
      }
      continue;
    }
    if(strcmp (Param,"InitIntron")==0) {
      iintron = strtod (File_Name,NULL);
      if  (errno == ERANGE) {
	fprintf (stderr, "ERROR:  Bad init intron parameter in config_file: %s\n", File_Name);
	exit(0);
      }
      continue;
    }
    if(strcmp (Param,"InitI0")==0) {
      iin[0] = strtod (File_Name,NULL);
      if  (errno == ERANGE) {
	fprintf (stderr, "ERROR:  Bad init intron 0 parameter in config_file: %s\n", File_Name);
	exit(0);
      }
      continue;
    }
    if(strcmp (Param,"InitI1")==0) {
      iin[1] = strtod (File_Name,NULL);
      if  (errno == ERANGE) {
	fprintf (stderr, "ERROR:  Bad init intron 1 parameter in config_file: %s\n", File_Name);
	exit(0);
      }
      continue;
    }
    if(strcmp (Param,"InitI2")==0) {
      iin[2] = strtod (File_Name,NULL);
      if  (errno == ERANGE) {
	fprintf (stderr, "ERROR:  Bad init intron 2 parameter in config_file: %s\n", File_Name);
	exit(0);
      }
      continue;
    }
    if(strcmp (Param,"MeanIntergen")==0) {
      minterg = strtod (File_Name,NULL);
      if  (errno == ERANGE) {
	fprintf (stderr, "ERROR:  Bad mean intergenic parameter in config_file: %s\n", File_Name);
	exit(0);
      }
      continue;
    }
    if(strcmp (Param,"Mean5'UTR")==0) {
      m5utr = strtod (File_Name,NULL);
      if  (errno == ERANGE) {
	fprintf (stderr, "ERROR:  Bad mean5'utr threshold score in config_file: %s\n", File_Name);
	exit(0);
      }
      continue;
    }
    if(strcmp (Param,"Mean3'UTR")==0) {
      m3utr = strtod (File_Name,NULL);
      if  (errno == ERANGE) {
	fprintf (stderr, "ERROR:  Bad mean3'utr threshold score in config_file: %s\n", File_Name);
	exit(0);
      }
      continue;
    }
    if(strcmp (Param,"MeanIntron")==0) {
      mintron = strtod (File_Name,NULL);
      if  (errno == ERANGE) {
	fprintf (stderr, "ERROR:  Bad meanintron threshold score in config_file: %s\n", File_Name);
	exit(0);
      }
      continue;
    }
    if(strcmp (Param,"PSngl")==0) {
      score = strtod (File_Name,NULL);
      score *= snglfactor;

      Trans[INTERG][ESGLPLUS] = 0.5 *score;
      if  (errno == ERANGE) {
	fprintf (stderr, "ERROR:  Bad sgl transition score in config_file: %s\n", File_Name);
	exit(0);
      }
      Trans[ESGLMINUS][INTERG]=Trans[INTERG][ESGLPLUS];
      Trans[INTERG][EINPLUS]=Trans[EINMINUS][INTERG]=0.5-Trans[INTERG][ESGLPLUS];
      continue;
    }
    if(strcmp (Param,"PTerm")==0) {
      Trans[I0PLUS][ETRPLUS] = strtod (File_Name,NULL);
      if  (errno == ERANGE) {
	fprintf (stderr, "ERROR:  Bad terminal transition score in config_file: %s\n", File_Name);
	exit(0);
      }
      Trans[ETRMINUS][I0MINUS]=Trans[ETRMINUS][I1MINUS]=Trans[ETRMINUS][I2MINUS]=Trans[I1PLUS][ETRPLUS]=Trans[I2PLUS][ETRPLUS]=Trans[I0PLUS][ETRPLUS];

      if(Use_Exon_Count) 
	Trans[E0MINUS][I0MINUS]=Trans[E1MINUS][I1MINUS]=Trans[E2MINUS][I2MINUS]=Trans[I0PLUS][E0PLUS]=Trans[I1PLUS][E1PLUS]=Trans[I2PLUS][E2PLUS]=1;
      else 
	Trans[E0MINUS][I0MINUS]=Trans[E1MINUS][I1MINUS]=Trans[E2MINUS][I2MINUS]=Trans[I0PLUS][E0PLUS]=Trans[I1PLUS][E1PLUS]=Trans[I2PLUS][E2PLUS]=1-Trans[I0PLUS][ETRPLUS];

      continue;
    }
    if(strcmp (Param,"PInitIn0")==0) {
      Trans[EINPLUS][I0PLUS]=strtod (File_Name,NULL);
      if  (errno == ERANGE) {
	fprintf (stderr, "ERROR:  Bad init-i0 transition score in config_file: %s\n", File_Name);
	exit(0);
      }
      Trans[I0MINUS][EINMINUS]=Trans[EINPLUS][I0PLUS];
      continue;
    }
    if(strcmp (Param,"PInitIn1")==0) {
      Trans[EINPLUS][I1PLUS]=strtod (File_Name,NULL);
      if  (errno == ERANGE) {
	fprintf (stderr, "ERROR:  Bad init-i1 transition score in config_file: %s\n", File_Name);
	exit(0);
      }
      Trans[I1MINUS][EINMINUS]=Trans[EINPLUS][I1PLUS];
      continue;
    }
    if(strcmp (Param,"PInitIn2")==0) {
      Trans[EINPLUS][I2PLUS]=strtod (File_Name,NULL);
      if  (errno == ERANGE) {
	fprintf (stderr, "ERROR:  Bad init-i2 transition threshold score in config_file: %s\n", File_Name);
	exit(0);
      }
      Trans[I2MINUS][EINMINUS]=Trans[EINPLUS][I2PLUS];
      continue;
    }
    if(strcmp (Param,"PIn0In0")==0) {
      Trans[E0PLUS][I0PLUS]=strtod (File_Name,NULL);
      if  (errno == ERANGE) {
	fprintf (stderr, "ERROR:  Bad i0-i0 transition score in config_file: %s\n", File_Name);
	exit(0);
      }
      Trans[I0MINUS][E0MINUS]=Trans[E0PLUS][I0PLUS];
      continue;
    }
    if(strcmp (Param,"PIn0In1")==0) {
      Trans[E0PLUS][I1PLUS]=strtod (File_Name,NULL);
      if  (errno == ERANGE) {
	fprintf (stderr, "ERROR:  Bad i0-i1 transition score in config_file: %s\n", File_Name);
	exit(0);
      }
      Trans[I1MINUS][E0MINUS]=Trans[E0PLUS][I1PLUS];  
      continue;
    }
    if(strcmp (Param,"PIn0In2")==0) {
      Trans[E0PLUS][I2PLUS]=strtod (File_Name,NULL);
      if  (errno == ERANGE) {
	fprintf (stderr, "ERROR:  Bad i0-i2 transition score in config_file: %s\n", File_Name);
	exit(0);
      }
      Trans[I2MINUS][E0MINUS]=Trans[E0PLUS][I2PLUS];
      continue;
    }
    if(strcmp (Param,"PIn1In0")==0) {
      Trans[E1PLUS][I0PLUS]=strtod (File_Name,NULL);
      if  (errno == ERANGE) {
	fprintf (stderr, "ERROR:  Bad i1-i0 transition score in config_file: %s\n", File_Name);
	exit(0);
      }
      Trans[I0MINUS][E1MINUS]=Trans[E1PLUS][I0PLUS];
      continue;
    }
    if(strcmp (Param,"PIn1In1")==0) {
      Trans[E1PLUS][I1PLUS]=strtod (File_Name,NULL);
      if  (errno == ERANGE) {
	fprintf (stderr, "ERROR:  Bad i1-i1 transition score in config_file: %s\n", File_Name);
	exit(0);
      }
      Trans[I1MINUS][E1MINUS]=Trans[E1PLUS][I1PLUS];
      continue;
    }
    if(strcmp (Param,"PIn1In2")==0) {
      Trans[E1PLUS][I2PLUS]=strtod (File_Name,NULL);
      if  (errno == ERANGE) {
	fprintf (stderr, "ERROR:  Bad i1-i2 transition score in config_file: %s\n", File_Name);
	exit(0);
      }
      Trans[I2MINUS][E1MINUS]=Trans[E1PLUS][I2PLUS];
      continue;
    }
    if(strcmp (Param,"PIn2In0")==0) {
      Trans[E2PLUS][I0PLUS]=strtod (File_Name,NULL);
      if  (errno == ERANGE) {
	fprintf (stderr, "ERROR:  Bad i2-i0 transition score in config_file: %s\n", File_Name);
	exit(0);
      }
      Trans[I0MINUS][E2MINUS]=Trans[E2PLUS][I0PLUS];
      continue;
    }
    if(strcmp (Param,"PIn2In1")==0) {
      Trans[E2PLUS][I1PLUS]=strtod (File_Name,NULL);
      if  (errno == ERANGE) {
	fprintf (stderr, "ERROR:  Bad i2-i1 transition score in config_file: %s\n", File_Name);
	exit(0);
      }
      Trans[I1MINUS][E2MINUS]=Trans[E2PLUS][I1PLUS];
      continue;
    }
    if(strcmp (Param,"PIn2In2")==0) {
      Trans[E2PLUS][I2PLUS]=strtod (File_Name,NULL);
      if  (errno == ERANGE) {
	fprintf (stderr, "ERROR:  Bad i2-i2 transition score in config_file: %s\n", File_Name);
	exit(0);
      }
      Trans[I2MINUS][E2MINUS]=Trans[E2PLUS][I2PLUS];
      continue;
    }
    if(strcmp (Param,"LengthDistrFile")==0) {
      strcpy(EXON_DISTR,File_Name);
      continue;
    }
    if(strcmp (Param,"ExonNoDistrFile")==0) {
      strcpy(EXONNO_DISTR,File_Name);
      continue;
    }
    if(strcmp(Param,"SignalPFile")==0) {
      strcpy(SIGNALP_FILE,File_Name);
      ifSP=1;
      continue;
    }
    if(strcmp(Param,"BoostExon")==0) {
      boostexon=atoi(File_Name);
      continue;
    }
    if(strcmp(Param,"BoostSgl")==0) {
      score = strtod (File_Name,NULL);
      if  (errno == ERANGE)
	fprintf (stderr, "ERROR:  Bad atg threshold score in config_file: %s\n", File_Name);
      else snglfactor=score;
      continue;
    }
  }
  fclose(fp);

  // Update constant transitions:

  Trans[ETRPLUS][INTERG]=Trans[INTERG][ETRMINUS]=1;
  Trans[INTERG][ESGLMINUS]=Trans[ESGLPLUS][INTERG]=1;


  // ---------------------------- INIT_DECISION_TREES ------------------

  // read atg/gt/ag/tag scores
  strcpy (File_Name, TRAIN_DIR);
  strcat (File_Name,"atg.dt");
  
  atgroot = (struct tree_node *) malloc (sizeof(struct tree_node));
  if ((atgroot = read_tree(File_Name)) == NULL) {
    fprintf(stderr,"Mktree: Cannot read atg.dt\n");
    exit(-1);	
  }

  strcpy (File_Name, TRAIN_DIR);
  strcat (File_Name,"acc.dt");
  
  accroot = (struct tree_node *) malloc (sizeof(struct tree_node));
  if ((accroot = read_tree(File_Name)) == NULL) {
    fprintf(stderr,"Mktree: Cannot read acc.dt\n");
    exit(-1);	
  }

  strcpy (File_Name, TRAIN_DIR);
  strcat (File_Name,"don.dt");
  
  donroot = (struct tree_node *) malloc (sizeof(struct tree_node));
  if ((donroot = read_tree(File_Name)) == NULL) {
    fprintf(stderr,"Mktree: Cannot read don.dt\n");
    exit(-1);	
  }

  strcpy (File_Name, TRAIN_DIR);
  strcat (File_Name,"stop.dt");
  
  tagroot = (struct tree_node *) malloc (sizeof(struct tree_node));
  if ((tagroot = read_tree(File_Name)) == NULL) {
    fprintf(stderr,"Mktree: Cannot read stop.dt\n");
    exit(-1);	
  }


  // ---------------------------- SCORE --------------------------------

  // the model file names should be in the parameter file
  strcpy (File_Name, TRAIN_DIR);
  strcat (File_Name, CODING_FILE);
   
  Read_Coding_Model (File_Name);

  // read the noncoding model
  strcpy (File_Name, TRAIN_DIR);
  strcat (File_Name, NONCODING_FILE);
  NMODEL=Read_NonCoding_Model(File_Name);

  // allocate space

  for(i=0;i<6;i++) {
   Cod[i]=(double *) malloc((Data_Len+2)*sizeof(double));
   if (Cod[i] == NULL) {
     fprintf(stderr,"Memory allocation for coding %d failure.\n",i);
     abort();
   }
   Cod[i][0]=0;
   Cod[i][Data_Len+1]=0;
  }

  for(i=0;i<2;i++) {
    SS[i]=(SpliceSite *) malloc((Data_Len+2)*sizeof(SpliceSite));
    if (SS[i] == NULL) {
      fprintf(stderr,"Memory allocation for splice sites failure.\n"); 
      abort();
    }
    SS[i][0].type=-1;
    SS[i][0].score=NO_SCORE;
    SS[i][Data_Len+1].type=-1;
    SS[i][Data_Len+1].score=NO_SCORE;
  }
  
  // do the actual scoring here

  /* forward strand */

  indexhex=0;
  for(i=1;i<Data_Len+1;i++) {

    // coding 

    // if sequence is in the same frame as Data starting at 1 then use Cod[0]
    // if sequence is in the same frame as Data starting at 2 then use Cod[1]
    // if sequence is in the same frame as Data starting at 3 then use Cod[2]

    if(i<MODEL_LEN) {
      noncoding = get_prob_of_window2(i-1,NMODEL,Data+1);
    }
    else {
      noncoding = get_prob_of_window1(i-MODEL_LEN+1,NMODEL,Data);
    }
    if(noncoding==0) noncoding=0.00001; // take a very low probability

    for(j=0;j<3;j++) 
      if(i>j) {

	if(i>=j+MODEL_LEN) {
	  model = (int)(i-MODEL_LEN-j)%3;
	  coding = get_prob_of_window1(i-MODEL_LEN+1,MODEL[model],Data);
	}
	else {
	  model = ((int)(i-j-1)%3 +1)%3;
	  coding = get_prob_of_window2(i-j-1,MODEL[model],Data+j+1);
	}
	if(coding==0) coding=0.00001; // take a very low probability

	Cod[j][i]=Cod[j][i-1]+log2(coding)-log2(noncoding);
	//Cod[j][i]=Cod[j][i-1]+0.5*log2(coding)-log2(noncoding); //if using DT

      }
      else { Cod[j][i]=0; }

    //printf("%d %c %f %f %f %f\n",i,Data[i],Cod[0][i],Cod[1][i],Cod[2][i],noncoding);

    
    // splice sites

    SS[0][i].type=-1;
    SS[0][i].score=NO_SCORE;

    /* atg's : 0 */

    if(i<Data_Len-1 && Data[i]=='a' && Data[i+1]=='t' && Data[i+2]=='g') { // Deal w/ start sites
      
      SS[0][i].type=0;
      
      if(i>12 && i<=Data_Len-6) {
	k=0;
	for(j=i-12;j<i+7;j++) {
	  switch (Data[j]){
	  case 'a': B[k]=0;break;
	  case 'c': B[k]=1;break;
	  case 'g': B[k]=2;break;
	  case 't': B[k]=3;break;
	  }
	  k++;
	}
	ret=Is_Atg(B,&score,Atg_Thr);
	SS[0][i].score=score;
      }
    }

    /* gt's : 1 */

    if(i<Data_Len && Data[i]=='g' && Data[i+1]=='t') { // Deal with donors
      
      SS[0][i].type=1;

      if(i>80 && i<=Data_Len-81) {
	k=0;
	for(j=i-80;j<i+82;j++){
	  switch (Data[j]){
	  case 'a': B[k]=0;break;
	  case 'c': B[k]=1;break;
	  case 'g': B[k]=2;break;
	  case 't': B[k]=3;break;
	  }
	  k++;
	}

	ret=Is_Donor(B,&score, Don_Thr, istdon);
	SS[0][i].score=score;
      }
    }

    /* ag's : 2 */

    if(i<Data_Len && Data[i]=='a' && Data[i+1]=='g') { // Deal with acceptors

      SS[0][i].type=2;

      if(i>80 && i<=Data_Len-81) {
	k=0;
	for(j=i-80;j<i+82;j++){
	  switch (Data[j]){
	  case 'a': B[k]=0;break;
	  case 'c': B[k]=1;break;
	  case 'g': B[k]=2;break;
	  case 't': B[k]=3;break;
	  }
	  k++;
	}

	ret=Is_Acceptor(B,&score, Acc_Thr, istacc);
	SS[0][i].score=score;
      }
    }

    /* stop codons : 3 */
    
    if( i < Data_Len-1 && 
       ((Data[i]=='t' && Data[i+1]=='a' && Data[i+2]=='a') ||
       (Data[i]=='t' && Data[i+1]=='g' && Data[i+2]=='a') ||
       (Data[i]=='t' && Data[i+1]=='a' && Data[i+2]=='g'))) {
      
      SS[0][i].type=3;

      if(i>4 && i<=Data_Len-14) {
	k=0;
	 for(j=i-4;j<i+15;j++) {
	   switch (Data[j]){
	   case 'a': B[k]=0;break;
	   case 'c': B[k]=1;break;
	   case 'g': B[k]=2;break;
	   case 't': B[k]=3;break;
	   }
	   k++;
	 }

	 ret=Is_Stop(B,&score,Stop_Thr);
	 SS[0][i].score=score;
      }
    }
  }

  // exit(0);

  /* reverse strand */

  char *Copy;

  Copy=(char *) malloc((Data_Len+2)*sizeof(char));
  if (Copy == NULL) {
    fprintf(stderr,"Memory allocation for copy failure.\n");
    abort();
  }
  

  strcpy(Copy,Data);
  Reverse_Complement(Copy,Data_Len);

  indexhex=0;
  for(i=1;i<Data_Len+1;i++) {

    // coding 

    if(i<MODEL_LEN) {
      noncoding = get_prob_of_window2(i-1,NMODEL,Copy+1);
    }
    else {
      noncoding = get_prob_of_window1(i-MODEL_LEN+1,NMODEL,Copy);
    }
    if(noncoding==0) noncoding=0.00001; // take a very low probability

    for(j=0;j<3;j++) 
      if(i>j) {

	if(i>=j+MODEL_LEN) {
	  model = (int)(i-MODEL_LEN-j)%3;
	  coding = get_prob_of_window1(i-MODEL_LEN+1,MODEL[model],Copy);
	}
	else {
	  model = ((int)(i-j-1)%3 +1)%3;
	  coding = get_prob_of_window2(i-j-1,MODEL[model],Copy+j+1);
	}
	if(coding==0) coding=0.00001; // take a very low probability

	Cod[3+j][i]=Cod[3+j][i-1]+log2(coding)-log2(noncoding);
	//Cod[3+j][i]=Cod[3+j][i-1]+0.5*log2(coding)-log2(noncoding); //if using DT

      }
      else { Cod[3+j][i]=0; }
    
    
    // splice sites

    SS[1][Data_Len-i+1].type=-1;
    SS[1][Data_Len-i+1].score=NO_SCORE;

    /* atg's : 0 */

    if(i<Data_Len-1 && Copy[i]=='a' && Copy[i+1]=='t' && Copy[i+2]=='g') { // Deal w/ start sites
      
      SS[1][Data_Len-i+1].type=0;
      
      if(i>12 && i<=Data_Len-6) {
	k=0;
	for(j=i-12;j<i+7;j++) {
	  switch (Copy[j]){
	  case 'a': B[k]=0;break;
	  case 'c': B[k]=1;break;
	  case 'g': B[k]=2;break;
	  case 't': B[k]=3;break;
	  }
	  k++;
	}
	ret=Is_Atg(B,&score,Atg_Thr);
	SS[1][Data_Len-i+1].score=score;
      }
    }

    /* gt's : 1 */

    if(i<Data_Len && Copy[i]=='g' && Copy[i+1]=='t') { // Deal with donors
      
      SS[1][Data_Len-i+1].type=1;

      if(i>80 && i<=Data_Len-81) {
	k=0;
	for(j=i-80;j<i+82;j++){
	  switch (Copy[j]){
	  case 'a': B[k]=0;break;
	  case 'c': B[k]=1;break;
	  case 'g': B[k]=2;break;
	  case 't': B[k]=3;break;
	  }
	  k++;
	}

	ret=Is_Donor(B,&score, Don_Thr, istdon);
	SS[1][Data_Len-i+1].score=score;
      }
    }

    /* ag's : 2 */

    if(i<Data_Len && Copy[i]=='a' && Copy[i+1]=='g') { // Deal with acceptors

      SS[1][Data_Len-i+1].type=2;

      if(i>80 && i<=Data_Len-81) {
	k=0;
	for(j=i-80;j<i+82;j++){
	  switch (Copy[j]){
	  case 'a': B[k]=0;break;
	  case 'c': B[k]=1;break;
	  case 'g': B[k]=2;break;
	  case 't': B[k]=3;break;
	  }
	  k++;
	}

	ret=Is_Acceptor(B,&score, Acc_Thr, istacc);
	SS[1][Data_Len-i+1].score=score;
      }
    }

    /* stop codons : 3 */
    
    if( i < Data_Len-1 && 
       ((Copy[i]=='t' && Copy[i+1]=='a' && Copy[i+2]=='a') ||
       (Copy[i]=='t' && Copy[i+1]=='g' && Copy[i+2]=='a') ||
       (Copy[i]=='t' && Copy[i+1]=='a' && Copy[i+2]=='g'))) {
      
      SS[1][Data_Len-i+1].type=3;

      if(i>4 && i<=Data_Len-14) {
	k=0;
	 for(j=i-4;j<i+15;j++) {
	   switch (Copy[j]){
	   case 'a': B[k]=0;break;
	   case 'c': B[k]=1;break;
	   case 'g': B[k]=2;break;
	   case 't': B[k]=3;break;
	   }
	   k++;
	 }

	 ret=Is_Stop(B,&score,Stop_Thr);
	 SS[1][Data_Len-i+1].score=score;
      }
    }
  }
  free(Copy);

  // load exon no distr file

  strcpy(File_Name,TRAIN_DIR);
  strcat(File_Name,EXONNO_DISTR);
  fp = fopen (File_Name, "r");
  if  (fp == NULL)
    {
      fprintf (stderr, "ERROR:  Unable to open file %s\n", File_Name);
      exit (0);
    }

  fgets (Line, MAX_LINE, fp);
  sscanf(Line,"%d",&ExNo);

  ExDistr=(double *) malloc((ExNo+1)*sizeof(double));
  if (ExDistr == NULL) {
    fprintf(stderr,"Memory allocation for no of exons distribution failure.\n");
    abort();
  }
  ExTail=(double *) malloc((ExNo+1)*sizeof(double));
  if (ExTail == NULL) {
    fprintf(stderr,"Memory allocation for tail no of exons distribution failure.\n");
    abort();
  }

  for(i=0;i<=ExNo;i++) {
    fscanf(fp,"%f",&readval);
    if (readval < 0.000001) readval = 0.000001;
    ExDistr[i]=readval;
  }

  fclose(fp);

  ExTail[0]=1;
  ExTail[ExNo]=ExDistr[ExNo];
  for(i=ExNo-1;i>0;i--) {
    ExTail[i]=ExDistr[i]+ExTail[i+1];
  }
  
  // load exon distr file

  strcpy(File_Name,TRAIN_DIR);
  strcat(File_Name,EXON_DISTR);
  fp = fopen (File_Name, "r");
  if  (fp == NULL)
    {
      fprintf (stderr, "ERROR:  Unable to open file %s\n", File_Name);
      exit (0);
    }
  
  while(fgets (Line, MAX_LINE, fp) != NULL) {
    if(strncmp(Line,"Initial",7)==0) {
      sscanf(Line,"%*s %d",&ret);
      LenDistr[0]=(double *) malloc(ret*sizeof(double));
      if (LenDistr[0] == NULL) {
	fprintf(stderr,"Memory allocation for initial exon length distribution failure.\n");
	abort();
      }
      LenNo[0]=ret;
      for(i=0;i<LenNo[0];i++) {
	fscanf(fp,"%f",&readval);
      if (readval < 0.000001) readval = 0.000001;
      LenDistr[0][i]=boostdistr+readval;
      }
    }
    else if(strncmp(Line,"Internal",8)==0) {
      sscanf(Line,"%*s %d",&ret);
      LenDistr[1]=(double *) malloc(ret*sizeof(double));
      if (LenDistr[1] == NULL) {
	fprintf(stderr,"Memory allocation for internal exon length distribution failure.\n");
	abort();
      }
      LenNo[1]=ret;
      for(i=0;i<LenNo[1];i++) {
	fscanf(fp,"%f",&readval);
      if (readval < 0.000001) readval = 0.000001;
      LenDistr[1][i]=boostdistr+readval;
      }
    }
    else if(strncmp(Line,"Terminal",8)==0) {
      sscanf(Line,"%*s %d",&ret);
      LenDistr[2]=(double *) malloc(ret*sizeof(double));
      if (LenDistr[2] == NULL) {
	fprintf(stderr,"Memory allocation for terminal exon length distribution failure.\n");
	abort();
      }
      LenNo[2]=ret;
      for(i=0;i<LenNo[2];i++) {
	fscanf(fp,"%f",&readval);
      if (readval < 0.000001) readval = 0.000001;
      LenDistr[2][i]=boostdistr+readval;
      }
    }
    else if(strncmp(Line,"Single",6)==0) {
      sscanf(Line,"%*s %d",&ret);
      LenDistr[3]=(double *) malloc(ret*sizeof(double));
      if (LenDistr[3] == NULL) {
	fprintf(stderr,"Memory allocation for single exon length distribution failure.\n");
	abort();
      }
      LenNo[3]=ret;
      for(i=0;i<LenNo[3];i++) {
	fscanf(fp,"%f",&readval);
      if (readval < 0.000001) readval = 0.000001;
      LenDistr[3][i]=boostdistr+readval;
      }
    }
  }

  fclose(fp);
  
  // now compute the geometric parameters
  for(i=4;i<6;i++) {
    LenDistr[i]=(double *) malloc(4*sizeof(double)); // first two are normal probabilities (1=stop; 2 =continue); 
    // last two are log probabilities
    if (LenDistr[i] == NULL) {
      fprintf(stderr,"Memory allocation for noncoding length distribution failure.\n");
      abort();
    }
  }
  
  // intron
  LenDistr[4][0]=1/(mintron+1);
  
  // intergenic
  LenDistr[5][0]=1/(minterg+m5utr+m3utr+1);

  for(i=4;i<6;i++) {
    LenDistr[i][1]=1-LenDistr[i][0];
    //LenDistr[i][2]=boostexon+log2(LenDistr[i][0]);
    LenDistr[i][2]=log2(LenDistr[i][0]);
    LenDistr[i][3]=log2(LenDistr[i][1]);
  }
    
  
  // load signalp matrix

  if(ifSP) { // there is a signalp file in the data
    strcpy(File_Name,TRAIN_DIR);
    strcat(File_Name,SIGNALP_FILE);
    fp = fopen (File_Name, "r");
    if  (fp == NULL)
      {
	fprintf (stderr, "ERROR:  Unable to open file %s\n", File_Name);
	exit (0);
      }
    for(i=0;i<32;i++) {
      for(j=0;j<4;j++) {
	fscanf(fp,"%d",&ret);
	SignalP[i][j]=ret;
      }
    }
  }


  // alloc sites memory and init sites

  iinterg+=i5utr+i3utr;

  for(val=INTERG; val<= I2MINUS; val++) {
    Sites[val] = (Site *) malloc((Data_Len+2)*sizeof(Site));
    if (Sites[val] == NULL) {
      fprintf(stderr,"Memory allocation for state sites failure.\n"); 
      abort();
    }

    switch(val) {
    case INTERG: Sites[val][0].score=0.5*log2(iinterg/iinterg);break;
    case I0PLUS:
    case I1PLUS:
    case I2PLUS: Sites[val][0].score=0.5*(log2(iintron/iinterg)+log2(iin[val-1]*3));break;
    case I0MINUS:
    case I1MINUS:
    case I2MINUS: Sites[val][0].score=0.5*(log2(iintron/iinterg)+log2(iin[val-4]*3));break;
    }
    Sites[val][0].prevex=NULL;
      
  }

  // process strand
  
  DP(1,Data_Len);

  // print data 

  maxscore=Sites[INTERG][Data_Len].score;

  maxval=INTERG;
  for(val=I0PLUS; val<=I2MINUS; val++) {

    if(Sites[val][Data_Len].score>maxscore) { 
      maxscore=Sites[val][Data_Len].score;
      maxval=val;
    }
  }

  curexon=Sites[maxval][Data_Len].prevex;
  
  PredExon=NULL;

  while(curexon!=NULL) {
    CurExon = (List *) malloc(sizeof(List));
    CurExon->predexon=curexon;
    CurExon->link=PredExon;
    PredExon=CurExon;
    curexon=curexon->prev;
  }

  printf("\nPredicted genes/exons\n\n");
  printf("Gene Exon Strand  Exon            Exon Range      Exon\n");
  printf("   #    #         Type                           Length\n\n");

  geneno=1;
  exonno=1;

  while(PredExon!=NULL) {
    printexon(PredExon->predexon,&geneno,&exonno);
    PredExon=PredExon->link;
  }

}

void printexon(exon *curexon,int *geneno,int *exonno) 
{
  switch(curexon->type) {
  case 0: printf("%4d %4d  +  Initial  %10ld %10ld  %7d\n",*geneno,*exonno,curexon->start,curexon->stop,curexon->stop-curexon->start+1);
    (*exonno)++;
    break;
  case 1: printf("%4d %4d  +  Internal %10ld %10ld  %7d\n",*geneno,*exonno,curexon->start,curexon->stop,curexon->stop-curexon->start+1);
    (*exonno)++;
    break;
  case 2: printf("%4d %4d  +  Terminal %10ld %10ld  %7d\n\n",*geneno,*exonno,curexon->start,curexon->stop,curexon->stop-curexon->start+1);
    (*geneno)++;*exonno=1;
    break;
  case 3: printf("%4d %4d  +  Single   %10ld %10ld  %7d\n\n",*geneno,*exonno,curexon->start,curexon->stop,curexon->stop-curexon->start+1);
    (*geneno)++;*exonno=1;
    break;
  case 4: printf("%4d %4d  -  Initial  %10ld %10ld  %7d\n\n",*geneno,*exonno,curexon->start,curexon->stop,curexon->stop-curexon->start+1);
    (*geneno)++;*exonno=1;
    break;
  case 5: printf("%4d %4d  -  Internal %10ld %10ld  %7d\n",*geneno,*exonno,curexon->start,curexon->stop,curexon->stop-curexon->start+1);
    (*exonno)++;
    break;
  case 6: printf("%4d %4d  -  Terminal %10ld %10ld  %7d\n",*geneno,*exonno,curexon->start,curexon->stop,curexon->stop-curexon->start+1);
    (*exonno)++;
    break;
  case 7: printf("%4d %4d  -  Single   %10ld %10ld  %7d\n\n",*geneno,*exonno,curexon->start,curexon->stop,curexon->stop-curexon->start+1);
    (*geneno)++;*exonno=1;
    break;
  }

}
   

exon *scoreinitexon(long int start,long int stop, int *prevstate, int *nextstate) 
{
  exon *newexon;
  POINT point;
  int len, codlen, frame, ind, lenin;
  double hexval,splicescore1,splicescore2;
  double codscore,transcore;
  double dts1,dts2;
  double SPimmscore,nSPimmscore;
  long int i,j;
  double sigp;

  newexon=(exon *) malloc(sizeof(exon));
  if(ifSP) sigp=score_signalp(start,stop);
  
  if(start<stop) { // forward init exon
    
    newexon->start=start;
    newexon->stop=stop;
    len=stop-start+1;
    newexon->lphase=0;
    newexon->rphase=len%3;
    newexon->exon_no=1;
  
    switch(newexon->rphase) {
    case 0: *nextstate=I0PLUS; break;
    case 1: *nextstate=I1PLUS; break;
    case 2: *nextstate=I2PLUS; break;
    }
    
    *prevstate=INTERG;

    newexon->type=0;
    newexon->prev=Sites[INTERG][start-1].prevex;
    
    transcore=0.5*log2(Trans[EINPLUS][*nextstate]);

    ind=0;
    
    splicescore1=SS[0][start].score;
    splicescore2=SS[0][stop+1].score;
  }
  else { // reverse init exon
    newexon->start=stop;
    newexon->stop=start;
    len=start-stop+1;
    newexon->lphase=len%3;
    newexon->rphase=0;

    switch(newexon->lphase) {
    case 0: *prevstate=I0MINUS; break;
    case 1: *prevstate=I1MINUS; break;
    case 2: *prevstate=I2MINUS; break;
    }

    *nextstate=INTERG;

    transcore=0.5*log2(Trans[*prevstate][EINMINUS]);

    newexon->type=4;
    newexon->prev=Sites[*prevstate][stop-1].prevex;

    if(newexon->prev) {
     newexon->exon_no=addexon(newexon->prev->exon_no);
     if(stopcodon(1,newexon->lphase,newexon->prev->stop,stop)) {
      free(newexon);
      return(NULL);
     }
     lenin=1+stop-newexon->prev->stop;
    }
    else {
      newexon->exon_no=2;
      lenin=stop-1;
    }

    transcore+=(double)lenin*(LenDistr[4][3]-LenDistr[5][3]);

    ind=1;

    splicescore2=SS[1][stop-1].score;
    splicescore1=SS[1][start].score;

    start=Data_Len-start+1;
    stop=Data_Len-stop+1;
  }

  // score splice sites
  point.dimension = (float *) malloc (sizeof(float));
  no_of_dimensions = 1;
  point.dimension[0] = splicescore1;
  classifysgl(&point, atgroot);
  dts1 = point.prob[0];
  free(point.dimension);

  point.dimension = (float *) malloc (sizeof(float));
  no_of_dimensions = 1;
  point.dimension[0] = splicescore2;
  classifysgl(&point, donroot);
  dts2 = point.prob[0];
  free(point.dimension);

#ifdef MESSAGE
  printf("\nsplicescore1=%f splicescore2=%f len=%d hexval=%f\n",splicescore1,splicescore2,len,hexval);
#endif

  frame=(start-1)%3;
  nSPimmscore=Cod[3*ind+frame][stop]-Cod[3*ind+frame][start-1];

  if(ifSP) {
    SPimmscore=sigp;
    if(len>63) SPimmscore+=Cod[3*ind+frame][stop]-Cod[3*ind+frame][start+62];
  }

  codlen=len/3;
  if(codlen>=LenNo[0]) codlen=LenNo[0]-1;

  if(ifSP) codscore = log2(0.2*pow(SPimmscore,2.0)+0.8*pow(nSPimmscore,2.0));
  else codscore = nSPimmscore;

  if(dts1==0) { dts1=0.000001;}
  if(dts1==1) { dts1=0.999999;}
  codscore += log2(dts1/(1-dts1));

  if(dts2==0) { dts2=0.000001;}
  if(dts2==1) { dts2=0.999999;}
  codscore += log2(dts2/(1-dts2));

  transcore += LenDistr[5][2] + log2(Trans[INTERG][EINPLUS]) - (double)len*LenDistr[5][3];


  newexon->score=boostexon+transcore+codscore+log2(LenDistr[0][codlen]);

#ifdef MESSAGE
  printf("initExon %ld %ld logpNP=%f interg->esgl+=%f logpNN=%f transcore=%f codscore=%f logdistr=%f totalscore=%f\n",newexon->start,newexon->stop,LenDistr[5][2],log2(Trans[INTERG][EINPLUS]),LenDistr[5][3],transcore,codscore,log2(LenDistr[0][codlen]),newexon->score);
#endif


  return(newexon);
}

exon *scoretermexon(long int start,long int stop, int *prevstate, int *nextstate) 
{
  exon *newexon;
  POINT point;
  int len, codlen, frame, ind, lenin;
  double hexval,splicescore1,splicescore2;
  double immscore,codscore,transcore;
  long int i,j;
  int lterm;
  double dts1,dts2;

  newexon=(exon *) malloc(sizeof(exon));
  
  if(start<stop) { // forward term exon
    newexon->start=start;
    newexon->stop=stop;
    len=stop-start+1;
    newexon->lphase=len%3;
    newexon->rphase=0;

    frame=newexon->lphase;

    switch(newexon->lphase) {
    case 0: *prevstate=I0PLUS; break;
    case 1: *prevstate=I2PLUS; break;
    case 2: *prevstate=I1PLUS; break;
    }
    
    *nextstate=INTERG;

    newexon->type=2;
    newexon->prev=Sites[*prevstate][start-1].prevex;

    if(newexon->prev) {
      newexon->exon_no=addexon(newexon->prev->exon_no);
      if(stopcodon(0,newexon->lphase,newexon->prev->stop,start)) {
        free(newexon);
        return(NULL);
      }
      lenin=1+start-newexon->prev->stop;
    }
    else {
      newexon->exon_no=2;
      lenin=start-1;
    }

    transcore=0.5*log2(Trans[EINPLUS][*prevstate]);
    
    transcore+=(double)lenin*(LenDistr[4][3]-LenDistr[5][3]);

    ind=0;

    splicescore1=SS[0][start-2].score;
    splicescore2=SS[0][stop-2].score;
  }
  else {
    newexon->start=stop;
    newexon->stop=start;
    len=start-stop+1;
    newexon->lphase=0;
    newexon->rphase=len%3;
    newexon->exon_no=1;

    frame=newexon->rphase;

    switch(newexon->rphase) {
    case 0: *nextstate=I0MINUS; break;
    case 1: *nextstate=I2MINUS; break;
    case 2: *nextstate=I1MINUS; break;
    }

    *prevstate=INTERG;

    transcore=0.5*log2(Trans[*nextstate][EINMINUS]);

    newexon->type=6;
    newexon->prev=Sites[INTERG][stop-1].prevex;

    ind=1;

    splicescore1=SS[1][start+2].score;
    splicescore2=SS[1][stop+2].score;

    start=Data_Len-start+1;
    stop=Data_Len-stop+1;
  }

  // score splice sites

  point.dimension = (float *) malloc (sizeof(float));
  no_of_dimensions = 1;
  point.dimension[0] = splicescore1;
  classifysgl(&point, accroot);
  dts1 = point.prob[0];
  free(point.dimension);

  point.dimension = (float *) malloc (sizeof(float));
  no_of_dimensions = 1;
  point.dimension[0] = splicescore2;
  classifysgl(&point, tagroot);
  dts2 = point.prob[0];
  free(point.dimension);

#ifdef MESSAGE
  printf("\n splicescore1=%f splicescore2=%f len=%d hexval=%f\n",splicescore1,splicescore2,len,hexval);
#endif
  
  frame+=start-1; // forward
  frame=frame%3;
  immscore=Cod[3*ind+frame][stop-3]-Cod[3*ind+frame][start-1]; // I took stop-3 to cancel the penalty for the stop codon;
                                                               // I hope it's not to important; I should score stops to avoid this
  
  codlen=len/3;
  if(codlen>=LenNo[2]) codlen=LenNo[2]-1;

  codscore = immscore;
  
  if(dts1==0) { dts1=0.000001;}
  if(dts1==1) { dts1=0.999999;}
  codscore += log2(dts1/(1-dts1));
  
  if(dts2==0) { dts2=0.000001;}
  if(dts2==1) { dts2=0.999999;}
  codscore += log2(dts2/(1-dts2));

  transcore += LenDistr[4][2]+log2(Trans[I0PLUS][ETRPLUS])- (double)len*LenDistr[5][3];

  newexon->score=boostexon+transcore+codscore+log2(LenDistr[2][codlen]);

#ifdef MESSAGE
  printf("termExon %ld %ld logpNP=%f intron->etr+=%f logpNN=%f transcore=%f codscore=%f logdistr=%f totalscore=%f\n",newexon->start,newexon->stop,LenDistr[4][2],log2(Trans[I0PLUS][ETRPLUS]),LenDistr[5][3],transcore,codscore,log2(LenDistr[2][codlen]),newexon->score);
#endif


  return(newexon);

}

exon *scoreinternexon(long int start,long int stop, int rphase, int *prevstate, int *nextstate) 
{
  exon *newexon;
  POINT point;
  int len, codlen, frame,ind, exonstate, lenin;
  double hexval,splicescore1,splicescore2;
  double immscore,codscore,transcore;
  long int i,j;
  double dts1,dts2;

  newexon=(exon *) malloc(sizeof(exon));
  newexon->rphase=rphase;

  if(start<stop) { // forward internal exon
    newexon->start=start;
    newexon->stop=stop;
    len=stop-start+1;
    newexon->lphase=(len-rphase)%3;

    frame=newexon->lphase;

    switch(newexon->lphase) {
    case 0: *prevstate=I0PLUS; exonstate=E0PLUS; break;
    case 1: *prevstate=I2PLUS; exonstate=E2PLUS; break;
    case 2: *prevstate=I1PLUS; exonstate=E1PLUS; break;
    }
    switch(rphase) {
    case 0: *nextstate=I0PLUS; break;
    case 1: *nextstate=I1PLUS; break;
    case 2: *nextstate=I2PLUS; break;
    }

    ind=0;

    newexon->type=1;
    newexon->prev=Sites[*prevstate][start-1].prevex;

    if(newexon->prev) {
     newexon->exon_no=addexon(newexon->prev->exon_no);  
     if(stopcodon(0,newexon->lphase,newexon->prev->stop,start)) {
      free(newexon);
      return(NULL);
     }
     lenin=1+start-newexon->prev->stop;
    }
    else {
      newexon->exon_no=2;
      lenin=start-1;
    }

    transcore = log2(Trans[exonstate][*nextstate]);

    transcore+=(double)lenin*(LenDistr[4][3]-LenDistr[5][3]);

    splicescore1=SS[0][start-2].score;
    splicescore2=SS[0][stop+1].score;
  }
  else {
    newexon->start=stop;
    newexon->stop=start;
    len=start-stop+1;
    newexon->lphase=(len-rphase)%3;
    frame=rphase;
    
    // reverse
    switch(newexon->lphase) {
    case 0: *prevstate=I0MINUS; break;
    case 1: *prevstate=I1MINUS; break;
    case 2: *prevstate=I2MINUS; break;
    }
    switch(rphase) {
    case 0: *nextstate=I0MINUS; exonstate=E0MINUS; break;
    case 1: *nextstate=I2MINUS; exonstate=E0MINUS; break;
    case 2: *nextstate=I1MINUS; exonstate=E0MINUS; break;
    }

    transcore = log2(Trans[*prevstate][exonstate]);

    newexon->type=5;
    newexon->prev=Sites[*prevstate][stop-1].prevex;

    if(newexon->prev) {
     newexon->exon_no=addexon(newexon->prev->exon_no);
     if(stopcodon(1,newexon->lphase,newexon->prev->stop,stop)) {
      free(newexon);
      return(NULL);
     }
     lenin=1+stop-newexon->prev->stop;
    }
    else {
      newexon->exon_no=2;
      lenin=stop-1;
    }

    transcore+=(double)lenin*(LenDistr[4][3]-LenDistr[5][3]);

    ind=1;

    splicescore1=SS[1][start+2].score;
    splicescore2=SS[1][stop-1].score;

    start=Data_Len-start+1;
    stop=Data_Len-stop+1;
  }

  // score splice sites  

  point.dimension = (float *) malloc (sizeof(float));
  no_of_dimensions = 1;
  point.dimension[0] = splicescore1;
  classifysgl(&point, accroot);
  dts1 = point.prob[0];
  free(point.dimension);

  point.dimension = (float *) malloc (sizeof(float));
  no_of_dimensions = 1;
  point.dimension[0] = splicescore2;
  classifysgl(&point, donroot);
  dts2 = point.prob[0];
  free(point.dimension);


#ifdef MESSAGE
  printf("\nsplicescore1=%f splicescore2=%f len=%d hexval=%f\n",splicescore1,splicescore2,len,hexval);
#endif

  frame+=start-1;
  frame=frame%3;

  immscore=Cod[3*ind+frame][stop]-Cod[3*ind+frame][start-1];

  codlen=len/3;
  if(codlen>=LenNo[1]) codlen=LenNo[1]-1;

  codscore = immscore;
  
  if(dts1==0) { dts1=0.000001;}
  if(dts1==1) { dts1=0.999999;}
  codscore += log2(dts1/(1-dts1));

  if(dts2==0) { dts2=0.000001;}
  if(dts2==1) { dts2=0.999999;}
  codscore += log2(dts2/(1-dts2));

  transcore += LenDistr[4][2]+log2(Trans[I0PLUS][E0PLUS])- (double)len*LenDistr[5][3];
  
  newexon->score=boostexon+transcore+codscore+log2(LenDistr[1][codlen]);

#ifdef  MESSAGE
  printf("internExon %ld %ld logpNP=%f intron->ex+=%f logpNN=%f transcore=%f codscore=%f logdistr=%f totalscore=%f lenin=%d introngain=%f\n",newexon->start,newexon->stop,LenDistr[4][2],log2(Trans[I0PLUS][E0PLUS]),LenDistr[5][3],transcore,codscore,log2(LenDistr[1][codlen]),newexon->score,lenin,LenDistr[4][3]-LenDistr[5][3]);
#endif

  return(newexon);
}

int check_for_stops(long int prevstop,long int newstart,int prevrphase)
{
  char stop[4];

  if(prevstop<newstart) { // check for forward stop codons at junction
    switch(prevrphase) {
    case 0: return(1);
    case 1: stop[0]=Data[prevstop]; stop[1]=Data[newstart]; stop[2]=Data[newstart+1]; break;
    case 2: stop[0]=Data[prevstop-1]; stop[1]=Data[prevstop]; stop[2]=Data[newstart]; break;
    }
    if((stop[0]=='t' && stop[1]=='a' && stop[2]=='a') ||
       (stop[0]=='t' && stop[1]=='g' && stop[2]=='a') ||
       (stop[0]=='t' && stop[1]=='a' && stop[2]=='g')) return(0);
  }
  else {
    switch(prevrphase) {
    case 0: return(1);
    case 1: stop[0]=Data[prevstop]; stop[1]=Data[newstart]; stop[2]=Data[newstart-1]; break;
    case 2: stop[0]=Data[prevstop+1]; stop[1]=Data[prevstop]; stop[2]=Data[newstart]; break;
    }
    if((stop[0]=='a' && stop[1]=='t' && stop[2]=='t') ||
       (stop[0]=='a' && stop[1]=='c' && stop[2]=='t') ||
       (stop[0]=='a' && stop[1]=='t' && stop[2]=='c')) return(0);
  }

  return(1);
}

exon *scoresglexon(long int start,long int stop) 
{
  exon *newexon;
  POINT point;
  int len, codlen, frame, ind;
  double hexval,splicescore1,splicescore2;
  double codscore, transcore;
  double SPimmscore,nSPimmscore;
  long int i,j;
  double sigp;
  double dts1,dts2;

  newexon=(exon *) malloc(sizeof(exon));
  newexon->lphase=0;
  newexon->rphase=0;
  newexon->exon_no=1;
  if(ifSP) sigp=score_signalp(start,stop);

  if(start<stop) { // forward sgl exon
    
    newexon->start=start;
    newexon->stop=stop;
    newexon->type=3;
    newexon->prev=Sites[INTERG][start-1].prevex;

    splicescore1=SS[0][start].score;
    splicescore2=SS[0][stop-2].score;

    ind=0;
  }
  else { // reverse sgl exon
    newexon->start=stop;
    newexon->stop=start;
    newexon->type=7;
    newexon->prev=Sites[INTERG][stop-1].prevex;
    
    splicescore1=SS[1][start].score;
    splicescore2=SS[1][stop+2].score;

    start=Data_Len-start+1;
    stop=Data_Len-stop+1;

    ind=1;
  }


  // score splice sites
  len=(int)(stop-start+1);

  point.dimension = (float *) malloc (sizeof(float));
  no_of_dimensions = 1;
  point.dimension[0] = splicescore1;
  classifysgl(&point, atgroot);
  dts1 = point.prob[0];
  free(point.dimension);

  point.dimension = (float *) malloc (sizeof(float));
  no_of_dimensions = 1;
  point.dimension[0] = splicescore2;
  classifysgl(&point, tagroot);
  dts2 = point.prob[0];
  free(point.dimension);


#ifdef MESSAGE
  printf("\nsplicescore1=%f splicescore2=%f len=%d hexval=%f\n",splicescore1,splicescore2,len,hexval);
#endif

  
  
  frame=(start-1)%3;
  nSPimmscore=Cod[3*ind+frame][stop-3]-Cod[3*ind+frame][start-1]; // I took stop-3 to cancel the penalty for the stop codon;
                                                               // I hope it's not to important; I should score stops to avoid this
  
  if(ifSP) {
    SPimmscore=sigp;
    if(len>66) SPimmscore+=Cod[3*ind+frame][stop-3]-Cod[3*ind+frame][start+62];
  }

  codlen=len/3;
  if(codlen>=LenNo[3]) codlen=LenNo[3]-1;

  if(ifSP) codscore = log2(0.2*pow(SPimmscore,2.0)+0.8*pow(nSPimmscore,2.0));
  else codscore = nSPimmscore;

  
  if(dts1==0) { dts1=0.000001;}
  if(dts1==1) { dts1=0.999999;}
  codscore += log2(dts1/(1-dts1));

  if(dts2==0) { dts2=0.000001;}
  if(dts2==1) { dts2=0.999999;}
  codscore += log2(dts2/(1-dts2));


  transcore= LenDistr[5][2] + log2(Trans[INTERG][ESGLPLUS]) - (double)len*LenDistr[5][3];

  newexon->score=boostexon+transcore+codscore+log2(LenDistr[3][codlen]);

#ifdef MESSAGE
  printf("sglExon %ld %ld logpNP=%f interg->esgl+=%f logpNN=%f transcore=%f codscore=%f logdistr=%f LenDistr[3][%d]=%f totalscore=%f\n",newexon->start,newexon->stop,LenDistr[5][2],log2(Trans[INTERG][ESGLPLUS]),LenDistr[5][3],transcore,codscore,log2(LenDistr[3][codlen]),codlen,LenDistr[3][codlen],newexon->score);  
#endif

  return(newexon);

}


void DP(long int start, long int stop) 
{
  long int i,j, thisstop;
  int s;
  long int laststop[6]; // remembers stops in all six reading frames
  exon *newexon;
  int prevstate,nextstate, rphase;
  double tempscore;


  for(i=0;i<6;i++) laststop[i]=0;

  for(i=start;i<=stop;i++) {

    if(i==stop) {
      printf("\n");
    }

    for(s=INTERG; s<= I2MINUS; s++) {
      Sites[s][i].score=Sites[s][i-1].score;
      Sites[s][i].prevex=Sites[s][i-1].prevex;
    }

    // forward strand
    
    if(i>3 && SS[0][i-3].type==3 && SS[0][i-3].score>=Stop_Thr) { // stop codon on forw strand at i-3,i-2,i-1
      
      // go back to score exons
      j=i-4;
      thisstop=laststop[(i-1)%3]-1;
      while(j>=1 && j>=thisstop) {

	if(SS[0][j].type==0 && (j+2)%3 == (i-1)%3 && SS[0][j].score>=Atg_Thr) { // single exon
	  newexon=scoresglexon(j,i-1);
	  tempscore = newexon->score + Sites[INTERG][j-1].score;
	  if(tempscore>Sites[INTERG][i].score) {


	    Sites[INTERG][i].score=tempscore;
	    if(Sites[INTERG][i].prevex != NULL && Sites[INTERG][i].prevex!=Sites[INTERG][i-1].prevex) 
	      free(Sites[INTERG][i].prevex);
	    Sites[INTERG][i].prevex=newexon;
	  }
	  else free(newexon);
	}

	if(SS[0][j].type==2 && SS[0][j].score >= Acc_Thr) { // terminal exon

	  newexon = scoretermexon(j+2,i-1,&prevstate,&nextstate);  

	  if(newexon) {

	    tempscore = newexon->score + Sites[prevstate][j+1].score;

	    if(Use_Exon_Count) { 
	      //tempscore += log2(ExDistr[newexon->exon_no])-log2(ExTail[newexon->exon_no]);
	      tempscore -= log2(ExTail[newexon->exon_no]);
	    }
	    if(tempscore>Sites[INTERG][i].score) {

	      //if(Use_Exon_Count) { tempscore-= log2(ExDistr[newexon->exon_no]);}
	      
	      Sites[INTERG][i].score=tempscore;
	      if(Sites[INTERG][i].prevex != NULL && Sites[INTERG][i].prevex!=Sites[INTERG][i-1].prevex) 
		free(Sites[INTERG][i].prevex);
	      Sites[INTERG][i].prevex=newexon;
	    }
	    else free(newexon);
	  }
	}

	j--;
      }

    }

    if(SS[0][i-3].type==3) laststop[(i-1)%3]=i-1;
    
    if(SS[0][i].type==1 && SS[0][i].score>=Don_Thr) { // GT: donor on forw strand at i,i+1

      // go back to score exons
      for(rphase=0;rphase<3;rphase++) {
	j=i-rphase-1;
	thisstop=laststop[(i-rphase-1)%3]-1;

	while(j>=1 && j>=thisstop) { 

	  if(SS[0][j].type==0 && j<=i-3-rphase && (j+2)%3 == (i-rphase-1)%3 && SS[0][j].score>=Atg_Thr) { // initial exon to Irphase+
	    newexon = scoreinitexon(j,i-1,&prevstate,&nextstate);



	    if(newexon) {

	      tempscore = newexon->score + Sites[INTERG][j-1].score;

	      if(Use_Exon_Count) { tempscore += log2(ExTail[2]);}

	      if(tempscore>Sites[nextstate][i].score) {
		Sites[nextstate][i].score=tempscore;
		if(Sites[nextstate][i].prevex != NULL && Sites[nextstate][i].prevex!=Sites[nextstate][i-1].prevex) 
		  free(Sites[nextstate][i].prevex);
		Sites[nextstate][i].prevex=newexon;
	      }
	      else free(newexon);
	    }
	  }

	  if(SS[0][j].type==2 && j+2<i-rphase-1 && SS[0][j].score>=Acc_Thr) { // internal exon


	    newexon = scoreinternexon(j+2,i-1,rphase,&prevstate,&nextstate);

	    if(newexon) {
	      tempscore=newexon->score+Sites[prevstate][j+1].score;

	      if(Use_Exon_Count) { tempscore += log2(ExTail[newexon->exon_no+1])-log2(ExTail[newexon->exon_no]);}
	      if(tempscore>Sites[nextstate][i].score && newexon->prev!=NULL && check_for_stops((newexon->prev)->stop,j+2,(newexon->prev)->rphase)) { // no stop codons at junction

		Sites[nextstate][i].score=tempscore;
		if(Sites[nextstate][i].prevex != NULL && Sites[nextstate][i].prevex!=Sites[nextstate][i-1].prevex) 
		  free(Sites[nextstate][i].prevex);
		Sites[nextstate][i].prevex=newexon;
		
	      }
	      else free(newexon);
	    }
	  }

	  j--;
	}

      }
    }

    // reverse strand
    if(SS[1][i-1].type==3) laststop[3+(i-1)%3]=i-3;

    if(SS[1][i-1].type==0 && SS[1][i-1].score>=Atg_Thr) { // start codon on rev strand at i-1, i-2, i-3
      
      // go back to score introns
      
      j=laststop[3+(i-1)%3];
      if(j>=1 && SS[1][j+2].score>=Stop_Thr) { // single exon
	newexon=scoresglexon(i-1,j);
	tempscore = newexon->score + Sites[INTERG][j-1].score;
	if(tempscore>Sites[INTERG][i].score) {

	  Sites[INTERG][i].score=tempscore;
	  if(Sites[INTERG][i].prevex != NULL && Sites[INTERG][i].prevex!=Sites[INTERG][i-1].prevex) 
	    free(Sites[INTERG][i].prevex);
	  Sites[INTERG][i].prevex=newexon;
	}
	else free(newexon);
      }

      thisstop=j+1;	
      j=i-4;
      while(j>=1 && j>=thisstop) {
	
	if(SS[1][j].type==1 && SS[1][j].score >= Don_Thr) { // initial exon
	  newexon = scoreinitexon(i-1,j+1,&prevstate,&nextstate);

	  if(newexon) {
	    tempscore = newexon->score + Sites[prevstate][j].score;

	    if(Use_Exon_Count) { 
	      //tempscore += log2(ExDistr[newexon->exon_no])-log2(ExTail[newexon->exon_no]);
	      tempscore -= log2(ExTail[newexon->exon_no]);
	    }
	    if(tempscore>Sites[INTERG][i].score) {

	      //if(Use_Exon_Count) { tempscore -= log2(ExDistr[newexon->exon_no]);}
	      Sites[INTERG][i].score=tempscore;
	      if(Sites[INTERG][i].prevex != NULL && Sites[INTERG][i].prevex!=Sites[INTERG][i-1].prevex) 
		free(Sites[INTERG][i].prevex);
	      Sites[INTERG][i].prevex=newexon;
	    }
	    else free(newexon);
	  }
	}

	j--;
      }
    }


    if(SS[1][i+1].type==2 && SS[1][i+1].score>=Acc_Thr) { // AG: acceptor on rev strand at i+1,i
      
      // go back to score exons
      for(rphase=0;rphase<3;rphase++) {
	j=i-2-rphase;
	thisstop=laststop[3+(i-1-rphase)%3]+1;
	
	if(thisstop>=2 && SS[1][thisstop+1].score>=Stop_Thr) {
	  newexon=scoretermexon(i-1,thisstop-1,&prevstate,&nextstate);  // first score terminal exon

	  if(newexon) {
	    tempscore = newexon->score + Sites[INTERG][thisstop-2].score;

	    if(Use_Exon_Count) { tempscore += log2(ExTail[2]);}

	    if(tempscore>Sites[nextstate][i].score) {

	      Sites[nextstate][i].score=tempscore;
	      if(Sites[nextstate][i].prevex != NULL && Sites[nextstate][i].prevex!=Sites[nextstate][i-1].prevex) 
		free(Sites[nextstate][i].prevex);
	      Sites[nextstate][i].prevex=newexon;
	    }
	    else free(newexon);
	  }
	}
	  
	while(j>=1 && j>=thisstop) {
	    
	  if(SS[1][j].type==1 && SS[1][j].score>=Don_Thr) { // internal donor
	    newexon = scoreinternexon(i-1,j+1,rphase, &prevstate, &nextstate);

	    if(newexon) {
	      tempscore=newexon->score+Sites[prevstate][j].score;

	      if(Use_Exon_Count) { tempscore += log2(ExTail[newexon->exon_no+1])-log2(ExTail[newexon->exon_no]);}
	      if(tempscore>Sites[nextstate][i].score && newexon->prev!=NULL && check_for_stops(j+1,(newexon->prev)->stop,newexon->lphase)) {

		Sites[nextstate][i].score=tempscore;
		if(Sites[nextstate][i].prevex != NULL && Sites[nextstate][i].prevex!=Sites[nextstate][i-1].prevex) 
		  free(Sites[nextstate][i].prevex);
		Sites[nextstate][i].prevex=newexon;
	      }
	      else free(newexon);
	    }
	  }
	  
	  j--;
	}
      }
    }
  }

}
  

void  Read_Coding_Model  (char * Param)

//  Read in the probability model indicated by  Param .

{
  FILE  * fp;
  
  fp = File_Open (Param, "r");   // maybe rb ?
  
  Read_Scoring_Model (fp);
  
  fclose (fp);
  
  return;
}

tModel *  Read_NonCoding_Model  (char * Param)

//  Read in the probability models of the non-coding regions.

{
  tModel * Delta;
  FILE  * fp;
  
  fp = File_Open (Param, "r");   // maybe rb ?
  
  Delta=Read_NonScoring_Model (fp);
  
  fclose (fp);

  return(Delta);
}


void  Process_Options  (int argc, char * argv [])
  
//  Process command-line options and set corresponding global switches
//  and parameters.
//
//    -d dir Set the training files' directory
//    -g n   Set minimum gene length to n
//    -r     Don't use independent probability score column
//    +r     Use independent probability score column
//    -t n   Set threshold score for calling as gene to n.  If the in-frame
//           score >= n, then the region is given a number and considered
//           a potential gene.
//    -f     Don't use filtering of the splice sites
//    +f     Use filtering of the splice sites
//    -e     Don't use exon count distribution
//    +e     Use exon count distribution (default)
//    -s     Don't use start site model
//    +s     Use start site model
//    -b     Don't use stop site model
//    +b     Use stop site model
//    -5 th  Use threshold th for the acceptor sites
//    -3 th  Use threshold th for the donor sites

  {
   char  * P;
   long int  W;
   double D;
   int  i;

   for  (i = 2;  i < argc;  i ++)
     {
      switch  (argv [i] [0])
        {
         case  '-' :
           switch  (argv [i] [1])
             {
	     case 'd' :   //  the training files' directory
	       strcpy(TRAIN_DIR,argv[++i]);strcat(TRAIN_DIR,"/");
	       //printf("train dir = %s\n",TRAIN_DIR);fflush(stdout);
	       break;
	     case  'g' :       // minimum gene length
	       errno = 0;
	       if  (argv [i] [2] != '\0')
		 P = argv [i] + 2;
	       else
		 P = argv [++ i];
	       W = strtol (P, NULL, 10);
	       if  (errno == ERANGE)
                    fprintf (stderr, "ERROR:  Bad minimum gene length %s\n", P);
	       else
		 Min_Gene_Len = W;
	       assert (Min_Gene_Len > 0);
	       break;
	     case  'r' :       // don't use random/independent score column
	       Use_Independent = FALSE;
	       break;
	     case  'e' :       // don't use exon count distribution; instead use distribution over number of exons as induced by GHMM
	       Use_Exon_Count = FALSE;
	       break;
	     case 'f' :
	       Use_Filter=0;
	       break;
	     case 's' :
	       Use_Start_Site=0;
	       break;
	       /*case  't' :       // threshold score for calling as gene
		 errno = 0;
		 if  (argv [i] [2] != '\0')
		 P = argv [i] + 2;
		 else
		 P = argv [++ i];
		 W = strtol (P, NULL, 10);
		 if  (errno == ERANGE)
		 fprintf (stderr, "ERROR:  Bad threshold score for gene %s\n", P);
		 else
		 Threshold_Score = W;
		 assert (Threshold_Score > 0 && Threshold_Score < 100);
		 break;*/
	     case '5':
	       errno = 0;
	       if  (argv [i] [2] != '\0')
		 P = argv [i] + 2;
	       else
		 P = argv [++ i];
	       D = strtod (P, NULL);
	       if  (errno == ERANGE)
		 fprintf (stderr, "ERROR:  Bad acceptor threshold score %s\n", P);
	       else
		 Acc_Thr = D;
	       break;
	     case '3':
	       errno = 0;
	       if  (argv [i] [2] != '\0')
		 P = argv [i] + 2;
	       else
		 P = argv [++ i];
	       D = strtod (P, NULL);
	       if  (errno == ERANGE)
		 fprintf (stderr, "ERROR:  Bad acceptor threshold score %s\n", P);
	       else
		 Don_Thr = D;
	       break;
	     default :
	       fprintf (stderr, "Unrecognized option %s\n", argv [i]);
             }
           break;
	case  '+' :
	  switch  (argv [i] [1])
	    {
	    case  'r' :       // use random/independent score column
	      Use_Independent = TRUE;
	      break;
	    case  'e' :       // don't use exon count distribution; 
	                      // instead use distribution over number of exons as induced by GHMM
	       Use_Exon_Count = TRUE;
	       break;
	    case 'f' :
	      Use_Filter = TRUE;
	      break;
	    case 's' :
	      Use_Start_Site = TRUE;
	      break;
	    default :
                fprintf (stderr, "Unrecognized option %s\n", argv [i]);
	    }
	  break;
	default :
	  fprintf (stderr, "Unrecognized option %s\n", argv [i]);
        }
     }

   return;
  }

void decision_tree()
{
  FILE *treenames;              /* file hold all trees' names */
  int i;
  char *s;
  char treefile[LINESIZE];
  char filename[550];
  char allfilename[550];

  struct tree_node
    **exontrees = NULL;       /* trees for classifing internal exons */
  struct tree_node
    **starttrees = NULL;       /* trees for classifying initial exons */
  struct tree_node
    **lasttrees = NULL;       /* trees for classifying final exons */
  struct tree_node
    **sngltrees = NULL;       /* trees for classifying single exons */


  double *if_6mervec;     /* in-frame hexamer freqs */

  /****************************************/
  /* read in the decision trees           */
  /****************************************/
  
  /* read in sets of trees for: 
     1. exons
     2. introns
     3. initial exons
     4. final exons  */
  
  /* allocate space for exonroots */
  exontrees = (struct tree_node **)
    malloc (sizeof(struct tree_node *) * no_of_trees);
  starttrees = (struct tree_node **)
    malloc (sizeof(struct tree_node *) * no_of_trees);
  lasttrees = (struct tree_node **)
    malloc (sizeof(struct tree_node *) * no_of_trees);
  sngltrees = (struct tree_node **)
    malloc (sizeof(struct tree_node *) * no_of_trees);

  /* the decision tree names must be in a file called "treenames" */
  strcpy(filename,TRAIN_DIR);
  strcat(filename,TREENAMES);
  if((treenames = fopen(filename, "r")) == NULL) {
    fprintf (stderr, "Decision tree names file %s can not be opened.",treenames);
    exit(-1);
  }
  for (i=0; i< (no_of_trees * 4); i++) {
    if ( fgets(treefile, LINESIZE, treenames) == NULL) {
      fprintf (stderr, "Not enough trees.\n");
      exit(-1);
    }
    s = strchr(treefile, '\n');
    *s = '\0';

    sprintf(filename,"%s%s",TRAIN_DIR,treefile);
    /* first 10 trees are for initial exons -- starttrees */
    if (i < 10) {
      //fprintf(stderr,"Read %s\n",treefile);
      if ((starttrees[i] = read_tree(filename)) == NULL) {
	fprintf(stderr,"Mktree: Cannot read %s.\n",treefile);
	exit(-1);	
      }
    }
    /* next 10 trees are for internal exons -- exontrees */
    else if (i < 20) {
      //fprintf(stderr,"Read %s\n",treefile);
      if ((exontrees[i-10] = read_tree(filename)) == NULL) {
	fprintf(stderr,"Mktree: Cannot read %s.\n",treefile);
	exit(-1);	
      }
    }
    /* next 10 trees are for final exons -- lasttrees */
    else if (i < 30) {
      //fprintf(stderr,"Read %s\n",treefile);
      if ((lasttrees[i-20] = read_tree(filename)) == NULL) {
	fprintf(stderr,"Mktree: Cannot read %s.\n",treefile);
	exit(-1);
      }
    }
    /* next 10 trees are for single exons */
    else if (i < 40) {
      //fprintf(stderr,"Read %s\n",treefile);
      if ((sngltrees[i-30] = read_tree(filename)) == NULL) {
	fprintf(stderr,"Mktree: Cannot read %s.\n",treefile);
	exit(-1);
      }
    }
  }
  fclose(treenames);

  exonroots = exontrees;
  startroots = starttrees;
  lastroots = lasttrees;
  snglroots = sngltrees;


  /****************************************/
  /* read in the hexamer frequency vector */
  /****************************************/

  if_6mervec = (double *) malloc (sizeof(double) * 4096); 
  //get_if_hexamer(IF_6MER_TRAIN, TRAIN_6MERS, 4096, if_6mervec);
  sprintf(filename,"%s%s",TRAIN_DIR,IF_6MER_TRAIN);
  sprintf(allfilename,"%s%s",TRAIN_DIR,TRAIN_6MERS);
  get_if_hexamer(filename, allfilename, 4096, if_6mervec);
  HexamerVec    = if_6mervec;

}


void classify (POINT *point,struct tree_node **roots)
{
  int j, t;
  struct tree_node *cur_node;
  double sum;
  double probtmp;

  /* initialize Exon and Intron prob */
  point->prob[0]=0;
  point->prob[1]=0;
  /* get length and skip if <= 2 */
  for (t=1; t<=no_of_trees; t++) {
    //printf("Tree %d\n",t);
    cur_node = roots[t-1];
    while (cur_node != NULL) {
      sum = cur_node->coefficients[no_of_dimensions];
      for (j=1;j<=no_of_dimensions;j++)
	sum += cur_node->coefficients[j-1] * point->dimension[j-1];
      if (sum < 0) {
	if (cur_node->left != NULL) 
	  cur_node = cur_node->left;
	else {
	  /* New for prob. classification, added by Xin Chen */
	  probtmp =
	    (double)cur_node->left_count[(cur_node->left_cat)-1]/
	    cur_node->left_total;
	  
	  if(cur_node->left_cat == 1) {
	    point->prob[0] += probtmp;
	    point->prob[1] += 1-probtmp;
	  }
	  else {
	    point->prob[0] += 1-probtmp;
	    point->prob[1] += probtmp;
	  }
	  
	  /****/
	  /*printf("left cat= %d  left_count[%d]= %d no_points= %d\n",
	     cur_node->left_cat,
	     cur_node->left_cat,
	     cur_node->left_count[cur_node->left_cat],
	     cur_node->left_total);*/
	  
	  break;
	}
      }
      else {
	if (cur_node->right != NULL) 
	  cur_node = cur_node->right;
	else {
	  /* New for prob. classification, added by Xin Chen */
	  probtmp =
	    (double)cur_node->right_count[(cur_node->right_cat)-1]/
	    cur_node->right_total;
	  
	  if(cur_node->right_cat == 1) {
	    point->prob[0] += probtmp;
	    point->prob[1] += 1-probtmp;
	  }
	  else {
	    point->prob[0] += 1-probtmp;
	    point->prob[1] += probtmp;
	  }
	  
	  /****/
	  /*printf("right cat= %d  right_count[%d]= %d no_points= %d\n",
	       cur_node->right_cat,
	       cur_node->right_cat,
	       cur_node->right_count[cur_node->right_cat],
	       cur_node->right_total);*/
	    
	  break;
	}
      }
    }
  }

  point->prob[0] /= (double) no_of_trees;
  point->prob[1] /= (double) no_of_trees;
  if(point->prob[0] >= point->prob[1]){
    point->category = 1;
  }
  else{
    point->category = 2;
  }
}


void classifysgl (POINT *point,struct tree_node *root)
{
  int j, t;
  double sum;
  double probtmp;
  struct tree_node *cur_node;

  /* initialize Exon and Intron prob */
  point->prob[0]=0;
  point->prob[1]=0;
  /* get length and skip if <= 2 */
  cur_node=root;
  while (cur_node != NULL) {
    sum = cur_node->coefficients[no_of_dimensions];
    for (j=1;j<=no_of_dimensions;j++)
      sum += cur_node->coefficients[j-1] * point->dimension[j-1];
    if (sum < 0) {
      if (cur_node->left != NULL) 
	cur_node = cur_node->left;
      else {
	/* New for prob. classification, added by Xin Chen */
	probtmp =
	  (double)cur_node->left_count[(cur_node->left_cat)-1]/
	  cur_node->left_total;
	
	if(cur_node->left_cat == 1) {
	  point->prob[0] += probtmp;
	  point->prob[1] += 1-probtmp;
	}
	else {
	  point->prob[0] += 1-probtmp;
	  point->prob[1] += probtmp;
	}
	
	/****/
	/*printf("left cat= %d  left_count[%d]= %d no_points= %d\n",
	  cur_node->left_cat,
	  cur_node->left_cat,
	  cur_node->left_count[cur_node->left_cat],
	  cur_node->left_total);*/
	
	break;
      }
    }
    else {
      if (cur_node->right != NULL) 
	cur_node = cur_node->right;
      else {
	/* New for prob. classification, added by Xin Chen */
	probtmp =
	  (double)cur_node->right_count[(cur_node->right_cat)-1]/
	  cur_node->right_total;
	
	if(cur_node->right_cat == 1) {
	  point->prob[0] += probtmp;
	  point->prob[1] += 1-probtmp;
	}
	else {
	  point->prob[0] += 1-probtmp;
	  point->prob[1] += probtmp;
	}
	
	/****/
	/*printf("right cat= %d  right_count[%d]= %d no_points= %d\n",
	  cur_node->right_cat,
	  cur_node->right_cat,
	  cur_node->right_count[cur_node->right_cat],
	  cur_node->right_total);*/
	
	break;
      }
    }
  }
  

  if(point->prob[0] >= point->prob[1]){
    point->category = 1;
  }
  else{
    point->category = 2;
  }
}


int get_if_hexamer(char *file1,char *file2,int nvals,double *vector)
{ 
   FILE *fp1;
   FILE *fp2;
   int index;
   double hexfreq;

   /* Open files.  The first file stores the in-frame hexamer
      frequencies.  The second file stores the hexamer frequencies
      for all frames of exons, introns, and intergenic DNA */
   if (((fp1=fopen(file1,"r")) == NULL)  ||
       ((fp2=fopen(file2,"r")) == NULL))
     {
       printf ("vector file -> %s\n", file1);
       fprintf(stderr,"Could not open one of %s or %s\n",file1,file2);
       exit(-1);
     }
   /* the file format is hexamer and frequency, e.g., 
      AGATCA  0.000245.  We just want to keep the numbers; in
      particular the log ratio between the frequencies. */
   for ( index=0; index<nvals; ++index )
     {
       fscanf(fp1,"%*s %lf",&(vector[index]));
       fscanf(fp2,"%*s %lf",&hexfreq);
       if (vector[index] != 0) /* avoid taking log of 0 */
	 vector[index] = log(vector[index]/hexfreq);
       else
	 vector[index] = -15.0;
     }
   fclose(fp1);
   fclose(fp2);
   return(0);
}


double ifhexamer(long int start, long int stop)
{ 
   int index,weight,val;                /* location of hexamer in HexamerVec */
   long int pos1, phase,i;
   static double *hex;            /* compute in 3 frames, return max */
   static int firsttime=1; 
   double retval;

   /* Initialize */
   if ( firsttime )
   {
      firsttime=0;
      hex = (double *) malloc( sizeof(double) * 3 );
   }

   if(start<stop) {
     for (phase=0; phase<3; phase++) {
       hex[phase] = 0;
       /* loop through the window in-phase, 3 at a time */
       for ( pos1=phase+start; pos1+5<stop+1; pos1 += 3 )
	 { 
	   index = 0;
	   weight = 1;
	   /* compute index into hexamerVec based on current hexamer */
	   for ( i=5; i>=0; --i )
	     {
	       /* pos1 is the current position in in the window */
	       switch(Data[pos1+i]) {
	       case 'a': val=0; break;
	       case 'c': val=1; break;
	       case 'g': val=2; break;
	       case 't': val=3; break;
	       default: val=1;
	       }
	       
	       index += weight*val;
	       weight *= 4;
	     }
	   /* we store "acgt" as "0123" and compute a unique number
	      based on this string, which is the index into HexamerVec.
	      HexamerVec stores log ratios, so we can add them. */
	   hex[phase] += HexamerVec[index];
	 }
     }
   }
   else {
     for (phase=0; phase<3; phase++) {
       hex[phase] = 0;
       /* loop through the window in-phase, 3 at a time */
       for ( pos1=start-phase; pos1-5>stop-1; pos1 -= 3 )
	 { 
	   index = 0;
	   weight = 1;
	   /* compute index into hexamerVec based on current hexamer */
	   for ( i=5; i>=0; --i )
	     {
	       /* pos1 is the current position in in the window */
	       switch(Data[pos1-i]) {
	       case 'a': val=3; break;
	       case 'c': val=2; break;
	       case 'g': val=1; break;
	       case 't': val=0; break;

		 /*case 'a': val=0; break;
	       case 'c': val=1; break;
	       case 'g': val=2; break;
	       case 't': val=3; break;*/

	       default: val=1;
	       }
	       
	       index += weight*val;
	       weight *= 4;
	     }
	   /* we store "acgt" as "0123" and compute a unique number
	      based on this string, which is the index into HexamerVec.
	      HexamerVec stores log ratios, so we can add them. */
	   hex[phase] += HexamerVec[index];
	 }
     }
   }
   printf("(hex[0]=%f hex[1]=%f hex[2]=%f) ",hex[0],hex[1],hex[2]);
   
   retval = hex[0];   /* return value, max of 3 phases */
   if (hex[1] > retval)
     retval = hex[1];
   if (hex[2] > retval)
     retval = hex[2];
   return(retval);
}

int basetoint(char c, int dir)
{
  if(dir==1) {
    switch(c) {
    case 'a':
    case 'A': return(0);
    case 'c':
    case 'C': return(1);
    case 'g':
    case 'G': return(2);
    case 't':
    case 'T': return(3);
    }
  }
  else {
    switch(c) {
    case 'a':
    case 'A': return(3);
    case 'c':
    case 'C': return(2);
    case 'g':
    case 'G': return(1);
    case 't':
    case 'T': return(0);
    }
  }

  return(-1);
}

double score_signalp(long int start, long int stop)
{
  int i,j,k;
  double score=0;
  int c1,c2,c3;

  j=0;
  if(start<stop) {
    for(k=start+3;k<stop-1 && k<=start+60;k+=3) {
      if(k==start+18) j=1;
      c1=basetoint(Data[k],1);
      c2=basetoint(Data[k+1],1);
      c3=basetoint(Data[k+2],1);
      i=c1*4+c2+16*j;
      score+=SignalP[i][c3];
    }
  }
  else {
    for(k=start-3;k>stop+1 && k>=start-60; k-=3) {
      if(k==start-18) j=1;
      c1=basetoint(Data[k],-1);
      c2=basetoint(Data[k-1],-1);
      c3=basetoint(Data[k-2],-1);
      i=c1*4+c2+16*j;
      score+=SignalP[i][c3];
    }
  }

  score/=10;

  return(score);
}

int stopcodon(int frame,int phase,int stop, int start)
{

  if(frame) { // reverse strand
    
    switch(phase) {
    case 0: return(0);
    case 1: 
      if((Data[start]=='a' && Data[stop]=='t' && Data[stop-1]=='c')||
	 (Data[start]=='a' && Data[stop]=='t' && Data[stop-1]=='t')||
	 (Data[start]=='a' && Data[stop]=='c' && Data[stop-1]=='t')) 
	return(1);
      else return(0);
    case 2:
      if((Data[start+1]=='a' && Data[start]=='t' && Data[stop]=='c')||
	 (Data[start+1]=='a' && Data[start]=='t' && Data[stop]=='t')||
	 (Data[start+1]=='a' && Data[start]=='c' && Data[stop]=='t')) 
	return(1);
      else return(0);
    }
  }
  else { // forward strand

    switch(phase) {
    case 0: return(0);
    case 1: 
      if((Data[stop-1]=='t' && Data[stop]=='a' && Data[start]=='g')||
	 (Data[stop-1]=='t' && Data[stop]=='a' && Data[start]=='a')||
	 (Data[stop-1]=='t' && Data[stop]=='g' && Data[start]=='a'))
	return(1);
      else return(0);
    case 2: 
      if((Data[stop]=='t' && Data[start]=='a' && Data[start+1]=='g')||
	 (Data[stop]=='t' && Data[start]=='a' && Data[start+1]=='a')||
	 (Data[stop]=='t' && Data[start]=='g' && Data[start+1]=='a'))
	return(1);
      else return(0);
    }
  }
}

int addexon(int number)
{
  int ret;

  ret = number+1 < ExNo ? number+1 : ExNo-1;

  return(ret);
}
